@@ -1,5 +1,6 @@
 /*
- * Copyright (c) 2004-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
+ * Copyright (c) 2004-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  *
  * This software is available to you under the terms of the
  * OpenIB.org BSD license included below:
@@ -34,9 +35,11 @@
 #include "agg_node.h"
 #include "agg_types.h"
 #include "am_bit_array.h"
+#include "am_common.h"
 #include "am_log.h"
+#include "common_types.h"
+#include "fabric.h"
 #include "fabric_dump.h"
-#include "job_manager.h"
 
 struct JobData
 {
@@ -50,6 +53,7 @@ struct JobData
     sharp_job_state job_state;
     std::vector<sharp_trees_t> tree_ids;
     std::vector<uint64_t> tree_feature_masks;
+    std::vector<uint8_t> trees_plane;
     uint16_t max_ost;
     uint32_t max_buffer;
     uint16_t max_groups;
@@ -61,7 +65,6 @@ struct JobData
     uint8_t num_channels_per_conn;
     uint8_t num_rails;
     uint8_t num_trees;
-    std::vector<port_key_t> an_port_keys;
     string job_file_path;
 };
 
@@ -94,11 +97,13 @@ class AggNodeFabricInfo : public AggNode
     AggNodeInfo m_all_available_quota_;     // For all jobs
 
     // DS for build trees
-    SetAnFabricInfoPtr m_descendant_leafs_;   // descendant leaf ANs
-    MapTreeTurns m_map_tree_turns_;           // count the number of identical tree
-                                              // turn in order to create unique trees
-    bool m_is_pure_ftree_;                    // true if only one path to
-                                              // every descendant.
+    // Non planarized table are located at m_descendant_leafs_ index 0
+    // Planarized tables are located according to plane index ( index = plane - 1)
+    VecSetAnFabricInfoPtr m_descendant_leafs_;   // descendant leaf ANs per plane
+    MapTreeTurns m_map_tree_turns_;              // count the number of identical tree
+                                                 // turn in order to create unique trees
+    bool m_is_pure_ftree_;                       // true if only one path to
+                                                 // every descendant.
 
     bool m_is_valid_;
     char m_date_str_[AM_DUMP_MAX_STR_LEN];
@@ -118,13 +123,15 @@ class AggNodeFabricInfo : public AggNode
     uint64_t
         m_priority_epoch_;   // This epoch specifies when last error happened. While picking trees for job, prefer AggNode with oldest error
 
+    MulticastForwardingData m_multicast_forwarding_data_;   // All-gather multicast trees forwarding table.
+
    public:
     AggNodeFabricInfo(const PortInfo& port_info,
                       const AggNodeInfo& agg_node_info,
                       sharp_an_id_t an_id,
                       uint64_t am_key,
-                      string switch_desc,
-                      const AnMinHopsTable an_min_hops_table,
+                      const string& switch_desc,
+                      const AnMinHopsTableByPlane& an_min_hops_table,
                       uint16_t num_llt_trees);
 
     const MapTreeIdToTreeNode& GetTreeNodes() const { return m_tree_nodes_; }
@@ -132,7 +139,7 @@ class AggNodeFabricInfo : public AggNode
     AMBitArray& GetAvailableTreeIdsLowPriority() { return m_available_tree_ids_low_priority; }
     void GetAvailableQuota(uint8_t priority, AggNodeInfo& available_quota) const;
 
-    const SetAnFabricInfoPtr& GetdescendantLeafs() const { return m_descendant_leafs_; }
+    SetAnFabricInfoPtr& GetdescendantLeafs(uint8_t plane);
 
     uint16_t GetMaxAllowedRadix() { return std::min(g_option_info.m_max_tree_radix, m_agg_node_info_.m_radix); }
 
@@ -150,9 +157,14 @@ class AggNodeFabricInfo : public AggNode
     }
     bool IsValid() const;
 
-    uint16_t GetTreesLoad() const { return m_trees_load_; }
+    // An AggNode can accpet MAD if it is Valid and not going through a phase of ISSU (FW Upgrade)
+    bool IsAcceptingMads() const { return IsValid() && (GetIssu() == SwitchIssuState::ISSU_STATE_IDLE); }
+    std::string AggNodeStateToString() const
+    {
+        return !IsValid() ? "invalid state" : ((GetIssu() != SwitchIssuState::ISSU_STATE_IDLE) ? "in middle of FW upgrade" : "valid");
+    }
 
-    bool IsRecoveryPending();
+    uint16_t GetTreesLoad() const { return m_trees_load_; }
 
     string TreesString() const
     {
@@ -213,7 +225,12 @@ class AggNodeFabricInfo : public AggNode
     int AllocQuota(const sharp_quota& quota, uint8_t priority);
     int FreeQuota(const Quota& quota, uint8_t priority);
 
-    void ClearDescendantLeafs() { m_descendant_leafs_.clear(); }
+    void ClearDescendantLeafs()
+    {
+        for (auto& descendant_leafs : m_descendant_leafs_) {
+            descendant_leafs.clear();
+        }
+    }
 
     void BuildDescendantLeafs();
     void DumpDescendants(FILE* f, const char* ident) const;
@@ -224,7 +241,8 @@ class AggNodeFabricInfo : public AggNode
     void GetSpanningInfo(const SetAnFabricInfoPtr& descendant_leafs,
                          ListSpanningInfo& spanning_info_list,
                          port_key_t parent_key,
-                         sharp_trees_t tree_id);
+                         sharp_trees_t tree_id,
+                         uint8_t plane);
 
     uint16_t GetTreeTurns(port_key_t parent_key, port_key_t descendant_key) const;
     void AddTreeTurn(port_key_t parent_key, port_key_t descendant_key);
@@ -239,7 +257,7 @@ class AggNodeFabricInfo : public AggNode
     inline uint16_t GetSatTreeLoad() const { return m_sat_tree_load_; }
     uint16_t GetSatJobLoadForSpecificTree(sharp_trees_t tree_id) const;   // There may be more than 1 job for each tree
     bool IsTreeNodeRoot(SharpJob* p_job, TreeNode* p_tree_node);
-    void RestoreFromNetworkExistingTreeNodes(int* p_operation_status, SetTreeIds tree_ids);
+    void RestoreFromNetworkExistingTreeNodes(int* p_operation_status, const SetTreeIds& tree_ids);
     void RestoreFromNetworkTreeToJobBindings(int* p_operation_status, const sharp_job_id_t sharp_job_id, const SetTreeIds tree_ids);
     void RestoreFromNetworkAllQps(int* p_operation_status);
     void RestoreFromNetworkExistingTreeNodesConfigurations(int* p_operation_status);
@@ -247,7 +265,7 @@ class AggNodeFabricInfo : public AggNode
     void RestoreFromNetworkAllJobIdsOfRunningJobs(int* p_operation_status);
     void RestoreFromNetworkAllJobQuotasOfRunningJobs(int* p_operation_status);
     void ReleaseQpsLeftFromSeamlessRestart();
-    void CleanDataRecoveredDuringSeamlessRestart();
+    void CleanNodeDataRecoveredDuringSeamlessRestart();
     void RestoreTreeNodeAggPaths(int& p_operation_status);
 
     MapRunningJobDataIter GetRunningJobsDataStartIter() { return m_running_job_data_.begin(); }
@@ -304,7 +322,7 @@ class AggNodeFabricInfo : public AggNode
     bool IsLockAvailable() const;
     bool IsExclusiveLockAvailable() const;
 
-    // Multicast
+    // LLT result Multicast
     uint32_t GetMulticastQpn();
 
     int AllocateMulticastQps(int* p_operation_status);
@@ -315,6 +333,12 @@ class AggNodeFabricInfo : public AggNode
 
     void UpdateLeafsSpanningLoad(uint16_t num_descendant_leafs) { m_leafs_spanning_load_ += num_descendant_leafs; }
 
+    // All gather multicast trees
+    MulticastForwardingData& GetMulticastData() { return m_multicast_forwarding_data_; }
+    int ConfigureMulticastTable(int* p_operation_status);
+    int ReadRequiredMulticastTables(ListLids& mlids, int* p_operation_status);
+    bool IsMlidConfiguredOnNode(lid_t mlid);
+
     ////////////////////////////////////////////
     /// DFP
     ////////////////////////////////////////////
@@ -322,10 +346,14 @@ class AggNodeFabricInfo : public AggNode
     uint8_t GetRootCount() { return m_root_count_; }
     void IncreaseRootCount() { m_root_count_++; }
 
-    uint32_t UpdateDescendantLeafs(const AnMinHopsTable& parent_an_min_hops_table, const SetAnFabricInfoPtr& parent_descendant_leafs);
+    uint32_t UpdateDescendantLeafs(const AnMinHopsTable& parent_an_min_hops_table,
+                                   const SetAnFabricInfoPtr& parent_descendant_leafs,
+                                   uint8_t plane);
 
-    uint8_t GetDfpTreeHeight();
-    uint16_t GetGroupNaiborsNumber() const;
+    uint8_t GetDfpTreeHeight(uint8_t plane);
+    uint16_t GetGroupNaiborsNumber(uint8_t plane) const;
+
+    void SetCleanOrUnconfigFailed(uint64_t epoch);
 
    private:
     QpStateEnum QPCConfigStateToQpStateEnum(int state)
@@ -344,13 +372,13 @@ class AggNodeFabricInfo : public AggNode
 
     void RecoverMulticastQps();
 
-    void AddDescendantLeafs(AggNodeFabricInfo* p_descendant_node);
+    void AddDescendantLeafs(AggNodeFabricInfo* p_descendant_node, uint8_t plane);
 
     // Get SetAnFabricInfoPtr of descendant that may span leafs
     // On DFP generates DescendantLeafs on all descendants
     // input parameter descendant_leafs is relevant only for DFP.
     //       get leafs that spans given descendant_leafs
-    void GetSpanningLeafs(SetAnFabricInfoPtr& spanning_leafs, const SetAnFabricInfoPtr& descendant_leafs);
+    void GetSpanningLeafs(SetAnFabricInfoPtr& spanning_leafs, const SetAnFabricInfoPtr& descendant_leafs, uint8_t plane);
 
     // call only after calling HasQuota
     void AllocQuota(AggNodeInfo& available_quota, const sharp_quota& quota);
@@ -392,6 +420,10 @@ class AggNodeFabricInfo : public AggNode
     void RemoveQpFromTreeCallback(FabricProviderCallbackContext* callback_context, int res, void* p_data);
 
     void ReleaseQpsCallback(FabricProviderCallbackContext* callback_context, int res, void* p_data);
+
+    void ConfigureMulticastTableCallback(FabricProviderCallbackContext* callback_context, int res, void* p_data);
+
+    void ReadRequiredMulticastTablesCallback(FabricProviderCallbackContext* callback_context, int res, void* p_data);
 };
 
 struct AnFabricInfoPtrSort
