@@ -39,6 +39,7 @@
 
 #include "agg_types.h"
 #include "am_common.h"
+#include "csv_parser.h"
 #include "option_manager.h"
 #include "port_data.h"
 
@@ -68,7 +69,8 @@ enum class SmdbPortState : uint8_t
     DOWN,
     INIT,
     ARM,
-    ACTIVE
+    ACTIVE,
+    ACTIVE_DEFER
 };
 
 enum class DiscThroughState : uint8_t
@@ -264,8 +266,12 @@ class SwitchRecord
     u_int8_t rank;
     u_int16_t coordinate;
     SwitchIssuState issu;
+    u_int16_t mlid_start;
+    u_int16_t mlid_end;
 
-    SwitchRecord() : node_guid(0), num_ports(0), status(0), rank(0), coordinate(0), issu(SwitchIssuState::ISSU_STATE_IDLE) {}
+    SwitchRecord()
+        : node_guid(0), num_ports(0), status(0), rank(0), coordinate(0), issu(SwitchIssuState::ISSU_STATE_IDLE), mlid_start(0), mlid_end(0)
+    {}
 
     static int Init(vector<ParseFieldInfo<class SwitchRecord>>& parse_section_info);
 
@@ -280,6 +286,10 @@ class SwitchRecord
     bool SetCoordinate(const char* field_str) { return CsvParser::Parse(field_str, coordinate, 16); }
 
     bool SetIssu(const char* field_str);
+
+    bool SetMlidStart(const char* field_str) { return CsvParser::Parse(field_str, mlid_start, 10); }
+
+    bool SetMlidEnd(const char* field_str) { return CsvParser::Parse(field_str, mlid_end, 10); }
 };
 
 class AnToAnRecord
@@ -505,7 +515,7 @@ class Port
          const PortTimestamp& timestamp,
          SpecialPortType special_port_type,
          uint16_t port_rate,
-         string port_label,
+         const string& port_label,
          u_int16_t aport,
          u_int8_t plane_number,
          u_int8_t number_of_planes)
@@ -544,6 +554,7 @@ class Port
     bool IsActive() const { return m_is_active_; }
     void Activate() { m_is_active_ = true; }
     void Disable() { m_is_active_ = false; }
+    inline bool IsStateChangeFromCurrentEpoch() const { return m_port_epoch_.m_curr_epoch_ == m_port_epoch_.m_prev_epoch_; }
 
     // get the port name
     string GetName() const;
@@ -696,7 +707,9 @@ class Node
     uint8_t m_first_plane_;
     uint8_t m_last_plane_;
     uint8_t m_max_plane_;
-    SwitchIssuState m_issu;
+    SwitchIssuState m_issu_;
+    uint16_t m_mlid_start_;
+    uint16_t m_mlid_end_;
 
     NodeTopologyInfo m_topology_info_;   // Topology information for this node.
 
@@ -710,7 +723,8 @@ class Node
 
     PrivateAppData m_app_data_1;
 
-    uint64_t m_node_epoch_;   // last update number
+    uint64_t m_node_epoch_;           // last update number
+    uint64_t m_state_change_epoch_;   // epoch when state was last changed
 
     // temp protocol variables
     mutable SetNodesPtr m_adjacent_switch_nodes_;
@@ -735,6 +749,9 @@ class Node
     inline NodeState GetPreviousState() const { return m_previous_state_; }
     inline void SetEpoch(uint64_t epoch) { m_node_epoch_ = epoch; }
     inline uint64_t GetCurrEpoch() const { return m_node_epoch_; }
+    inline uint64_t GetStateChangeEpoch() const { return m_state_change_epoch_; }
+    inline void SetStateChangeEpoch(uint64_t epoch) { m_state_change_epoch_ = epoch; }
+    inline bool IsStateChangeFromCurrentEpoch() const { return m_node_epoch_ == m_state_change_epoch_; }
     // get the node name
     inline const string& GetDescription() const { return m_description_; }
     inline NodeType GetType() const { return m_type_; }
@@ -762,13 +779,22 @@ class Node
     uint8_t GetPlaneLast() const { return m_last_plane_; }
     void SetMaxPlane(const uint8_t max_planes) { m_max_plane_ = max_planes; }
     uint8_t GetMaxPlane() const { return m_max_plane_; }
-    bool IsPlanarized() const { return m_first_plane_; }
-    bool IsPlanarizedMultiPlane() const { return m_first_plane_ != m_last_plane_; }
+    bool IsPlanarized() const
+    {   // Planarized with single plane like black mamba switch with AN per plane
+        return m_first_plane_ && m_last_plane_;
+    }
+    bool IsPlanarizedMultiPlane() const
+    {   // Planarized with multiple planes like crocodile switch with single AN for all planes
+        return m_first_plane_ && m_last_plane_ && m_first_plane_ != m_last_plane_;
+    }
     bool IsOnPlane(uint8_t plane) { return m_first_plane_ <= plane && plane <= m_last_plane_; }
     std::string PlaneToString() const;
-    void SetIssuState(SwitchIssuState issu) { m_issu = issu; };
-    SwitchIssuState GetIssuState() { return m_issu; }
-
+    void SetIssuState(SwitchIssuState issu) { m_issu_ = issu; };
+    SwitchIssuState GetIssuState() { return m_issu_; }
+    void SetNodeMlidStart(uint16_t mlid_start) { m_mlid_start_ = mlid_start; }
+    void SetNodeMlidEnd(uint16_t mlid_end) { m_mlid_end_ = mlid_end; }
+    uint16_t GetNodeMlidStart() { return m_mlid_start_; }
+    uint16_t GetNodeMlidEnd() { return m_mlid_end_; }
     void SetState(const NodeState new_state);
     string GetName() const;
 
@@ -813,6 +839,8 @@ class Node
 
     const SetNodesPtr& GetAdjacentSwitchNodes() const;
 
+    void LogNodePlaneMismatch(std::set<std::tuple<std::string, std::string, int, int>>& unique_pairs);
+
    private:
     void DumpDescLine(ostream& sout) const;
     bool UpdatesMinHopsTable(MinHopsTable& adjacent_min_hops_table, MinHopsIndexTable& adjacent_min_hops_index_table, uint8_t plane_index);
