@@ -1,5 +1,6 @@
 /*
- * Copyright (c) 2012-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
+ * Copyright (c) 2012-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -41,7 +42,6 @@
 #include <vector>
 // #include <fstream>
 #include <inttypes.h>
-#include <string.h>
 #include <sys/types.h>
 #include <chrono>
 #include <iomanip>
@@ -91,15 +91,22 @@ using std::vector;
 #define INVALID_TREE_RANK  0xFF
 #define INVALID_COORDINATE 0xFFFF
 
+#define INVALID_LID 0
+
 #define DEFAULT_AM_KEY 0
 
-#define MAX_TREES_PER_JOB         4
-#define MAX_USER_DATA_PER_OST     1024
-#define MAX_DFP_RANK              3
+#define MAX_TREES_PER_JOB     8
+#define MAX_USER_DATA_PER_OST 1024
+#define MAX_DFP_RANK          3
 
 #define MAX_CONTROL_PATH_VERSION_SUPPORTED 2
 
-#define DEFAULT_PLANE 1
+#define MIN_SHARP_MLID_START    0xC020
+#define MAX_SHARP_MLID_END      0xFFFF
+
+#define NON_PLANARIZED         ((uint8_t)0)
+#define PLANARIZED_FIRST_PLANE ((uint8_t)1)
+#define INVALID_PLANE          0xFF
 
 #ifndef PRIx64
 #if __WORDSIZE == 64
@@ -115,6 +122,8 @@ using std::vector;
 #define U64H_FMT "0x%016lx"
 #endif   // U64H_FMT
 
+#define UINT64_STR_FMT "0x" << std::setfill('0') << std::setw(16) << std::hex
+
 #ifndef U64D_FMT
 #define U64D_FMT "%lu"
 #endif   // U64D_FMT
@@ -134,11 +143,22 @@ using std::vector;
 
 #define MAX_SEMAPHORES 0x40
 
+#define MAX_PORTS_NUM 256
+
+// PLFT_ID 0 - SM
+// PLFT_ID 1 - AM, mapped to switch down ports (compute direction)
+// PLFT_ID 2 - AM, mapped to switch up ports (core sw direction)
+#define MAX_MULTICAST_PLFT_ID 3
+
+#define MULTICAST_TREES_DUMP_FILE "_mc_trees.mc_dump"
+
 ///////////////////////////////////////////////////////////////////////////////
 //
 // TYPEDEFS
 //
-struct PortKey;
+struct HcaPortKey;
+struct MulticastTableEntryI;
+struct MulticastTableUpdate;
 
 typedef uint8_t phys_port_t;
 typedef uint8_t child_index_t;
@@ -146,7 +166,10 @@ typedef uint8_t rank_t;
 typedef uint16_t lid_t;
 typedef uint16_t device_id_t;
 typedef uint16_t coordinate_t;
-typedef PortKey port_key_t;
+typedef uint16_t island_id_t;
+typedef uint16_t node_degree_t;
+typedef HcaPortKey hca_port_key_t;
+typedef uint64_t port_key_t;
 typedef uint32_t sharp_job_id_t;
 typedef uint64_t sharpd_id_t;
 typedef uint16_t sharp_trees_t;
@@ -159,18 +182,32 @@ typedef uint16_t sharp_an_id_t;
 typedef uint32_t sharp_path_id_t;
 typedef float sharp_percent_t;
 
-typedef std::vector<uint8_t> MinHopsTable;         // min hops table, key=LID, value=hops [0xFF, 0xFF, 1, 0, 0xFF, 2]
-typedef std::vector<uint16_t> MinHopsIndexTable;   // contain all keys(LID) of min hops table [2,3,5]
+typedef std::vector<uint8_t> MinHopsTable;                         // min hops table, key=LID, value=hops [0xFF, 0xFF, 1, 0, 0xFF, 2]
+typedef std::vector<uint16_t> MinHopsIndexTable;                   // contain all keys(LID) of min hops table [2,3,5]
+typedef std::vector<MinHopsTable> MinHopsTableByPlane;             // min hops table, key=LID, value=hops [0xFF, 0xFF, 1, 0, 0xFF, 2]
+typedef std::vector<MinHopsIndexTable> MinHopsIndexTableByPlane;   // contain all keys(LID) of min hops table [2,3,5]
 typedef lid_t node_min_hop_key_t;
 
 typedef std::set<class AggNodeFabricInfo*> SetAnFabricInfoPtr;
+typedef std::vector<SetAnFabricInfoPtr> VecSetAnFabricInfoPtr;
 typedef std::set<AggNodeFabricInfo*> SetAggNodeFabricPtr;   // Important to keep as ordered set, since we compare sets
 typedef std::list<class TreeNode*> ListTreeNodePtr;
 typedef std::vector<uint8_t> AnMinHopsTable;
+typedef std::vector<AnMinHopsTable> AnMinHopsTableByPlane;
 typedef std::map<phys_port_t, uint8_t> MapPortSemaphores;
 typedef std::map<phys_port_t, uint32_t> MapPortSATQps;
 typedef std::set<sharp_trees_t> SetTreeIds;
 typedef std::vector<sharp_trees_t> VecTreeIds;
+typedef std::list<sharp_trees_t> ListTreeIds;
+
+using ListLids = std::list<lid_t>;
+using SetPhysPorts = std::set<phys_port_t>;
+
+using MulticastTablePlftEntry = std::array<MulticastTableEntryI*, MAX_MULTICAST_PLFT_ID>;
+using MulticastTableEntryVec = std::vector<MulticastTablePlftEntry>;
+// MulticastForwardingTable dimensions are: [mlid_block][port_group][plft]
+using MulticastForwardingTable = std::vector<MulticastTableEntryVec>;
+using MulticastTableUpdates = std::set<MulticastTableUpdate>;
 
 // messages of agg node states
 extern char const* const g_removed_from_network_msg;
@@ -221,7 +258,6 @@ enum SeamlessRestartStatus
     PARSE_AGG_NODES_DUMP_FAILED,
     JOB_FILES_OLD_OR_CORRUPTED,
     NUMBER_OF_RETRIES_EXCEEDED_MAX_RETRIES,
-    AM_VERSION_DIFFERENT_FROM_DUMP_VERSION,
 
     // Seamless restart failed
     AGG_NODE_DUMP_NOT_FOUND,
@@ -238,9 +274,18 @@ enum SeamlessRestartStatus
     AGG_NODES_CONFIGURATON_COMPARE_FAILED,
     RESTORE_AGG_PATH_ACTION_FAILED,
     HANDLE_PENDING_TREE_EDGES_FAILED,
-    RESTORE_AGG_PATHS_FAILED
+    RESTORE_AGG_PATHS_FAILED,
+    QUERY_MULTICAST_TABLES_FAILED
 };
 
+enum class SwitchIssuState : uint8_t
+{
+    ISSU_STATE_IDLE = 0,         // Idle means the switch is working as usuall
+    ISSU_STATE_IN_PROGRESS = 1   // Issu process means the switch is working, but cannot handle MADs
+};
+
+const std::string SwitchIssuStateToString(const SwitchIssuState issu);
+
 static inline const char* SharpTrapNumberToChar(const SharpTrapNumberEnum trap_number)
 {
     switch (trap_number) {
@@ -288,29 +333,26 @@ struct Epoch
     }
 };
 
-struct PortKey
+struct HcaPortKey
 {
     uint64_t m_key;
     uint8_t m_plane;
 
-    PortKey() : m_key(0), m_plane(0) {}
-    PortKey(uint64_t key, uint8_t plane) : m_key(key), m_plane(plane) {}
-    uint64_t GetKey() const { return m_key; }
-    uint8_t GetPlane() const { return m_plane; }
-    std::string ToString() const { return std::to_string(m_key) + ':' + std::to_string(m_plane); }
+    HcaPortKey() : m_key(0), m_plane(0) {}
+    HcaPortKey(uint64_t key, uint8_t plane) : m_key(key), m_plane(plane){};
 
-    bool operator<(const PortKey& other) const
+    uint64_t GetKey() const { return m_key; }
+    std::string ToString() const;
+    bool operator<(const HcaPortKey& other) const
     {
-        if (m_key < other.m_key) {
-            return true;
-        } else if (m_key == other.m_key) {
+        if (m_key == other.m_key) {
             return m_plane < other.m_plane;
         }
-        return false;
+        return m_key < other.m_key;
     }
-    bool operator!=(const PortKey& other) const { return !(*this == other); }
+    bool operator!=(const HcaPortKey& other) const { return !(*this == other); }
 
-    bool operator==(const PortKey& other) const { return m_key == other.m_key && m_plane == other.m_plane; }
+    bool operator==(const HcaPortKey& other) const { return m_key == other.m_key && m_plane == other.m_plane; }
 };
 
 // needed for using PortKey as unordered_map/unordered_set key.
@@ -319,14 +361,9 @@ struct PortKey
 namespace std
 {
 template <>
-struct hash<PortKey>
+struct hash<HcaPortKey>
 {
-    std::size_t operator()(const PortKey& port_key) const noexcept
-    {
-        std::size_t h1 = std::hash<uint64_t>{}(port_key.m_key);
-        std::size_t h2 = std::hash<uint8_t>{}(port_key.m_plane);
-        return (h1 ^ (h2 << 1)) >> 1;
-    }
+    std::size_t operator()(const HcaPortKey& port_key) const noexcept { return std::hash<uint64_t>{}(port_key.m_key); }
 };
 }   // namespace std
 
@@ -397,6 +434,7 @@ struct JobResource
     uint16_t m_pkey;
     SharpExclusiveLock m_exclusive_lock;
     bool m_request_rmc;
+    uint8_t m_max_mc_trees;
 
     JobResource()
         : m_quota(),
@@ -409,10 +447,12 @@ struct JobResource
           m_req_feature_mask(0),
           m_pkey(0),
           m_exclusive_lock(NO_EXCLUSIVE_LOCK),
-          m_request_rmc(false)
+          m_request_rmc(false),
+          m_max_mc_trees(0)
     {}
 };
 
+// TopologyType must be synchnonized with src/smx/smx_swig/smx_types.py SharpTopologyType
 enum TopologyType
 {
     TOPOLOGY_TYPE_NONE,
@@ -610,7 +650,10 @@ struct PortInfo
     NodeTopologyInfo m_peer_topology_info;
     PortTimestamp m_timestamp;
     bool m_use_grh;
+    uint8_t m_first_plane;
+    uint8_t m_last_plane;
     uint8_t m_max_plane;
+    phys_port_t m_peer_port_num;
 
     // PortInfo() : m_port_key(0), m_lid(0), m_peer_key(0) {}
 
@@ -622,7 +665,10 @@ struct PortInfo
              port_key_t peer_key,
              NodeTopologyInfo peer_topology_info,
              const PortTimestamp& timestamp,
+             uint8_t first_plane,
+             uint8_t last_plane,
              uint8_t max_plane,
+             phys_port_t peer_port_num,
              bool use_grh = false)
         : m_port_key(port_key),
           m_subnet_prefix(subnet_prefix),
@@ -633,30 +679,35 @@ struct PortInfo
           m_peer_topology_info(peer_topology_info),
           m_timestamp(timestamp),
           m_use_grh(use_grh),
-          m_max_plane(max_plane)
+          m_first_plane(first_plane),
+          m_last_plane(last_plane),
+          m_max_plane(max_plane),
+          m_peer_port_num(peer_port_num)
     {}
 
     string GetName() const
     {
+        std::string pln = "";
+        if (IsPlanarized()) {
+            pln = " PLN:" + PlaneToString();
+        }
+
         std::stringstream stream;
         stream << m_node_desc;
-        if (m_max_plane > 1) {
-            stream << " PLN:" << std::to_string(m_port_key.GetPlane()) << " GUID:0x" << std::hex << m_port_key.GetKey()
-                   << " lid:" << std::dec << m_lid;
-        } else {
-            stream << " GUID:0x" << std::hex << m_port_key.GetKey();
-        }
+        stream << " GUID:0x" << std::hex << m_port_key << pln;
         return stream.str();
     }
 
     string GetSwString() const
     {
-        std::stringstream stream;
-        if (m_max_plane > 1) {
-            stream << "switch GUID:0x" << hex << m_peer_key.GetKey() << " PLN:" << std::to_string(m_peer_key.GetPlane());
-        } else {
-            stream << "switch GUID:0x" << hex << m_peer_key.GetKey();
+        std::string pln = "";
+        if (IsPlanarized()) {
+            pln = " PLN:" + PlaneToString();
         }
+
+        std::stringstream stream;
+        stream << "switch GUID:0x" << std::hex << m_peer_key << pln;
+
         return stream.str();
     }
 
@@ -668,17 +719,36 @@ struct PortInfo
             lmc_str = std::string(" lmc:") + std::to_string(m_lmc);
         }
         stream << m_node_desc;
-        if (m_max_plane > 1) {
-            stream << " PLN:" << std::to_string(m_port_key.GetPlane());
+        if (IsPlanarized()) {
+            stream << " PLN:" << PlaneToString();
         }
-        stream << " GUID:0x" << hex << m_port_key.GetKey() << " PEER GUID:0x" << m_peer_key.GetKey() << " lid:" << dec << m_lid << lmc_str
-               << " " << m_peer_topology_info.ToString();
+        stream << " GUID:0x" << hex << m_port_key << " PEER GUID:0x" << m_peer_key << " lid:" << dec << m_lid << lmc_str << " "
+               << m_peer_topology_info.ToString();
 
         return stream.str();
     }
 
     port_key_t GetPortKey() const { return m_port_key; }
     port_key_t GetPeerPortKey() const { return m_peer_key; }
+    // planarized data path return lid according to lmc\plane in crocodile switch
+    lid_t GetDataLid(uint8_t plane) const { return plane && IsOnPlane(plane) && m_lmc ? m_lid + plane - 1 : m_lid; }
+    bool IsInLidRange(lid_t lid) const
+    {
+        // return (lid & ~((0x1 << m_lmc) - 1)) == m_lid;
+
+        return (m_lid <= lid) && (lid <= (m_lid + (0x1 << m_lmc) - 1));   // lid < x < lid + 2^lmc-1
+    };
+    bool IsOnPlane(uint8_t plane) const { return m_first_plane <= plane && plane <= m_last_plane; }
+    bool IsPlanarizedMultiPlane() const
+    {   // Planarized with multiple planes like crocodile switch with single AN for all planes
+        return m_first_plane && m_last_plane && m_first_plane != m_last_plane;
+    };
+    bool IsPlanarizedSinglePlane() const
+    {   // Planarized with single plane like black mamba switch with AN per plane
+        return m_first_plane && m_first_plane == m_last_plane;
+    };
+    bool IsPlanarized() const { return m_first_plane && m_last_plane; };
+    std::string PlaneToString() const;
 };
 
 struct AggPortInfo
@@ -991,4 +1061,42 @@ struct hash<SharpExtJobId>
 };
 }   // namespace std
 
+struct MulticastTableEntryI
+{
+};
+
+struct MulticastTableUpdate
+{
+    uint16_t m_mlid_block;
+    uint8_t m_port_group;
+    uint8_t m_plft;
+
+    MulticastTableUpdate(uint16_t mlid_block, uint8_t port_group, uint8_t plft)
+        : m_mlid_block(mlid_block), m_port_group(port_group), m_plft(plft)
+    {}
+
+    bool operator<(const MulticastTableUpdate& other) const
+    {
+        if (m_mlid_block < other.m_mlid_block) {
+            return true;
+        }
+        if (m_mlid_block > other.m_mlid_block) {
+            return false;
+        }
+        if (m_port_group < other.m_port_group) {
+            return true;
+        }
+        if (m_port_group > other.m_port_group) {
+            return false;
+        }
+        return (m_plft < other.m_plft);
+    }
+};
+
+struct MulticastForwardingData
+{
+    MulticastForwardingTable m_table;
+    MulticastTableUpdates m_pending_updates;
+};
+
 #endif   // AGG_TYPES_H_
