@@ -27,8 +27,9 @@
 #define NVSHMEM_IBGDA_SUPPORT
 /* #undef NVSHMEM_IBGDA_SUPPORT_GPUMEM_ONLY */
 /* #undef NVSHMEM_ENABLE_ALL_DEVICE_INLINING */
+/* #undef NVSHMEM_HOSTLIB_ONLY */
 
-#if defined __clang_llvm_bitcode_lib__ || defined NVSHMEM_BITCODE_APPLICATION
+#if defined __clang_llvm_bitcode_lib__ || defined NVSHMEM_HOSTLIB_ONLY
 #undef NVSHMEM_IBGDA_SUPPORT
 #undef NVSHMEM_IBGDA_SUPPORT_GPUMEM_ONLY
 #define NVSHMEM_ENABLE_ALL_DEVICE_INLINING
